<?php

namespace Drupal\eca_content\Plugin\Action;

use Drupal\Core\Access\AccessibleInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\TypedData\ComplexDataInterface;
use Drupal\Core\TypedData\DataReferenceInterface;
use Drupal\Core\TypedData\ListInterface;
use Drupal\eca\Plugin\Action\ActionBase;
use Drupal\eca\Service\Conditions;
use Drupal\user\Entity\User;

/**
 * Replaces Drupal\Core\Field\FieldUpdateActionBase.
 *
 * We need to replace the core base class because within the ECA context
 * entities should not be saved after modifying a field value.
 *
 * The replacement is achieved with PHP's class_alias(), see eca_content.module
 */
abstract class FieldUpdateActionBase extends ActionBase {

  /**
   * Gets an array of values to be set.
   *
   * @return array
   *   Array of values with field names as keys.
   */
  abstract protected function getFieldsToUpdate();

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
      'load_entity_into_token' => TRUE,
      'load_current_user_into_token' => TRUE,
      'save_entity' => FALSE,
      'strip_tags' => FALSE,
      'trim' => FALSE,
    ] + parent::defaultConfiguration();
  }

  /**
   * Helper function to save the entity only outside ECA context or when
   * requested explicitly.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity which might have to be saved.
   */
  protected function save(ContentEntityInterface $entity): void {
    if (empty($entity->eca_context) || ($this->configuration['save_entity'] === Conditions::OPTION_YES)) {
      $entity->save();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function execute($entity = NULL) {
    /** @var \Drupal\Core\Entity\EntityInterface $entity */
    if ($entity === NULL) {
      return;
    }
    if ($this->configuration['load_entity_into_token'] === Conditions::OPTION_YES) {
      $this->tokenServices->addTokenData($entity->getEntityTypeId(), $entity);
    }
    if ($this->configuration['load_current_user_into_token'] === Conditions::OPTION_YES) {
      $this->tokenServices->addTokenData('user', User::load($this->currentUser->id()));
    }
    $all_entities_to_save = [];
    foreach ($this->getFieldsToUpdate() as $field => $value) {
      $output = $this->tokenServices->replace($value);
      if ($this->configuration['strip_tags'] === Conditions::OPTION_YES) {
        $output = preg_replace('/[\t\n\r\0\x0B]/', '', strip_tags($output));
      }
      if ($this->configuration['trim'] === Conditions::OPTION_YES) {
        $output = trim($output);
      }
      $entities_to_save = [];
      if ($update_target = $this->getUpdateTarget($entity, $field, $entities_to_save)) {
        // Try to set the value. If that attempt fails, then it would throw an
        // exception, and the exception would be logged as an error.
        $update_target->setValue($output);
        if (empty($entities_to_save)) {
          throw new \LogicException(sprintf("The provided field %s does not resolve for entities to be saved from the %s entity having ID %s.", $field, $entity->getEntityTypeId(), $entity->id()));
        }
        foreach ($entities_to_save as $entity_to_save) {
          if (!in_array($entity_to_save, $all_entities_to_save, TRUE)) {
            $all_entities_to_save[] = $entity_to_save;
          }
        }
      }
      else {
        // The property path does not exist, thus the provided field is not a
        // valid argument.
        throw new \InvalidArgumentException(sprintf("The provided field %s does not exist as a property path on the %s entity having ID %s.", $field, $entity->getEntityTypeId(), $entity->id()));
      }
    }
    foreach ($all_entities_to_save as $to_save) {
      $this->save($to_save);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function access($object, AccountInterface $account = NULL, $return_as_object = FALSE) {
    /** @var \Drupal\Core\Entity\EntityInterface $entity */
    $entity = $object;
    // @todo Try to find a simpler solution path for access logic.
    // @see https://www.drupal.org/project/drupal/issues/3244585
    $entity_op = 'update';
    $field_op = 'edit';

    /** @var \Drupal\Core\Access\AccessResultInterface $result */
    $result = $entity->access($entity_op, $account, TRUE);

    foreach (array_keys($this->getFieldsToUpdate()) as $field) {
      $entities_to_save = [];
      if ($update_target = $this->getUpdateTarget($entity, $field, $entities_to_save)) {
        if (empty($entities_to_save)) {
          throw new \LogicException(sprintf("The provided field %s does not resolve for entities to be saved from the %s entity having ID %s.", $field, $entity->getEntityTypeId(), $entity->id()));
        }
        foreach ($entities_to_save as $entity_to_save) {
          if ($entity_to_save !== $entity) {
            $result->andIf($entity_to_save->access($entity_op, $account, TRUE));
          }
        }
        $parent = $update_target->getParent();
        if ($parent instanceof AccessibleInterface) {
          $result->andIf($parent->access($field_op, $account, TRUE));
          $parent = $parent->getParent();
        }
      }
      else {
        throw new \InvalidArgumentException(sprintf("The provided field %s does not exist as a property path on the %s entity having ID %s.", $field, $entity->getEntityTypeId(), $entity->id()));
      }
    }

    return $return_as_object ? $result : $result->isAllowed();
  }

  /**
   * Returns the target to update for a given entity and property path.
   *
   * For being able to update a specific property, we allow to access it
   * by using the dot notation, traversing along an existing property path.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The source entity.
   * @param string $property_path
   *   The property path.
   * @param array &$entities_to_save
   *   The entities that needs to be saved in order to persist the changes.
   *
   * @return \Drupal\Core\TypedData\TypedDataInterface|null
   *   The update target, or NULL if not found.
   */
  protected function getUpdateTarget(ContentEntityInterface $entity, string $property_path, array &$entities_to_save) {
    if (empty($property_path)) {
      // The source is always an entity. Within this action, it cannot be
      // updated without at least specifying a field name.
      $entities_to_save = [];
      return NULL;
    }

    $parts = explode('.', $property_path);
    $last_i = count($parts) - 1;
    $data = $entity->getFields(TRUE);
    foreach ($parts as $i => $property) {
      if ($data instanceof ComplexDataInterface) {
        $data = $data->getProperties(TRUE);
      }
      if (($data instanceof \ArrayAccess || is_array($data)) && isset($data[$property])) {
        $data = $data[$property];
        if ($data instanceof DataReferenceInterface) {
          // Directly jump to the contained target, if any is present.
          // @todo Should entities be auto-created here?
          if (!($data = $data->getTarget())) {
            $entities_to_save = [];
            return NULL;
          }
        }
      }
      elseif ($data instanceof ListInterface) {
        // Allow for auto-appending an item, in case the given argument
        // indicates that it either does not care (i.e. the current property
        // is not a digit) or points to the next free delta in the list.
        if (!ctype_digit($property)) {
          // Some input may skip the delta, e.g. body.value is treated as an
          // equivalent to body.0.value.
          $data = $data->first() ?: $data->appendItem();
          if ($data instanceof ComplexDataInterface) {
            $data = $data->getProperties(TRUE);
            if (isset($data[$property])) {
              $data = $data[$property];
            }
            else {
              $entities_to_save = [];
              return NULL;
            }
          }
        }
        elseif ($property == count($data)) {
          $data = $data->appendItem();
        }
        else {
          $entities_to_save = [];
          return NULL;
        }
      }
      else {
        $entities_to_save = [];
        return NULL;
      }
      if ($i === $last_i) {
        // Scalar updates usually address a scalar target. In case the input has
        // not specified otherwise, use the main property. By going that way,
        // other existing property values are not getting cleared by accident.
        if ($data instanceof ListInterface) {
          $data = $data->first() ?: $data->appendItem();
        }
        if ($data instanceof ComplexDataInterface && ($main_property = $data->getDataDefinition()->getMainPropertyName())) {
          $data = $data->get($main_property);
        }
      }
      if (empty($entities_to_save) && ($value = $data->getValue()) && ($value instanceof EntityInterface) && ($value !== $entity) && $value->isNew()) {
        // During the traversal of the property path, a new entity has been
        // instantiated. Catch the first instantiated one and use it for saving.
        // The entity save operation chain then takes care for automatically
        // saving any other referenced entity which might be new too.
        $entities_to_save[] = $data->getValue();
      }
    }
    if (empty($entities_to_save)) {
      // So far, we have not yet catched a new entity that is a different object
      // than the source entity. This implies to save the one entity, which is
      // holding the addressed field.
      $root = $data->getRoot();
      if (method_exists($root, 'getEntity')) {
        $root = $root->getEntity();
        $entities_to_save[] = $root;
      }
    }
    return $data;
  }

}
