<?php

namespace Drupal\eca\Entity;

use Drupal\Component\EventDispatcher\Event;
use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\eca\Entity\Objects\EcaEvent;
use Drupal\eca\Entity\Objects\EcaGateway;
use Drupal\eca\Entity\Objects\EcaObject;
use Drupal\eca\Entity\Objects\EcaAction;
use Drupal\Core\Config\Entity\ConfigEntityBase;
use Drupal\eca\Plugin\ECA\Modeller\ModellerInterface;

/**
 * Defines the ECA entity type.
 *
 * @ConfigEntityType(
 *   id = "eca",
 *   label = @Translation("ECA"),
 *   label_collection = @Translation("ECAs"),
 *   label_singular = @Translation("ECA"),
 *   label_plural = @Translation("ECAs"),
 *   label_count = @PluralTranslation(
 *     singular = "@count ECA",
 *     plural = "@count ECAs",
 *   ),
 *   handlers = {
 *     "storage" = "Drupal\eca\Entity\EcaStorage",
 *     "list_builder" = "Drupal\eca\Entity\ListBuilder",
 *     "access" = "Drupal\eca\Entity\AccessControlHandler"
 *   },
 *   config_prefix = "eca",
 *   admin_permission = "administer eca",
 *   entity_keys = {
 *     "id" = "id",
 *     "label" = "label",
 *     "uuid" = "uuid",
 *     "status" = "status",
 *     "weight" = "weight"
 *   },
 *   config_export = {
 *     "id",
 *     "modeller",
 *     "label",
 *     "uuid",
 *     "status",
 *     "version",
 *     "weight",
 *     "events",
 *     "conditions",
 *     "gateways",
 *     "actions"
 *   }
 * )
 */
class Eca extends ConfigEntityBase {

  use EcaTrait;
  use EcaObjectTrait;

  /**
   * The ID.
   *
   * @var string
   */
  protected string $id;

  /**
   * The helpdesk label.
   *
   * @var string
   */
  protected string $label;

  /**
   * @var array
   */
  protected array $events;

  /**
   * @var array
   */
  protected array $conditions;

  /**
   * @var array
   */
  protected array $gateways;

  /**
   * @var array
   */
  protected array $actions;

  /**
   * @var \Drupal\eca\Entity\Model
   */
  protected Model $model;

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    parent::calculateDependencies();
    try {
      $this->addDependency('config', $this->getModel()->getConfigDependencyName());
    }
    catch (InvalidPluginDefinitionException | PluginNotFoundException $e) {
      // Ignroe this. Without a model, no dependencies.
    }
    return $this;
  }

  /**
   * @param $id
   *
   * @return string
   */
  protected function buildCacheId($id): string {
    return "eca:$this->id:$id";
  }

  /**
   * @return bool
   */
  public function isEditable(): bool {
    if ($modeller = $this->getModeller()) {
      return $modeller->isEditable();
    }
    return FALSE;
  }

  /**
   * @return \Drupal\eca\Plugin\ECA\Modeller\ModellerInterface|null
   */
  public function getModeller(): ?ModellerInterface {
    try {
      /** @var ModellerInterface $plugin */
      $plugin = $this->modellerPluginManager()->createInstance($this->get('modeller'));
    }
    catch (PluginException $e) {
      // Can be ignored.
    }
    if ($plugin !== NULL) {
      $plugin->setConfigEntity($this);
      return $plugin;
    }
    return NULL;
  }

  /**
   * @return \Drupal\eca\Entity\Model
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getModel(): Model {
    if (!isset($this->model)) {
      $storage = $this->entityTypeManager()->getStorage('eca_model');
      /** @var \Drupal\eca\Entity\Model $model */
      $model = $storage->load($this->id());
      if ($model === NULL) {
        $model = $storage->create([
          'id' => $this->id(),
        ]);
      }
      $this->model = $model;
    }
    return $this->model;
  }

  /**
   * @param string $id
   * @param string $plugin_id
   * @param array $fields
   *
   * @return $this
   */
  public function addCondition(string $id, string $plugin_id, array $fields): Eca {
    $this->conditions[$id] = [
      'plugin' => $plugin_id,
      'fields' => $fields,
    ];
    return $this;
  }

  /**
   * @param string $id
   * @param int $type
   * @param array $successors
   *
   * @return $this
   */
  public function addGateway(string $id, int $type, array $successors): Eca {
    $this->gateways[$id] = [
      'type' => $type,
      'successors' => $successors,
    ];
    return $this;
  }

  /**
   * @param string $id
   * @param string $plugin_id
   * @param string $label
   * @param array $fields
   * @param array $successors
   *
   * @return $this
   */
  public function addEvent(string $id, string $plugin_id, string $label, array $fields, array $successors): Eca {
    if (empty($label)) {
      $label = $id;
    }
    $this->events[$id] = [
      'plugin' => $plugin_id,
      'label' => $label,
      'fields' => $fields,
      'successors' => $successors,
    ];
    return $this;
  }

  /**
   * @param string $id
   * @param string $plugin_id
   * @param string $label
   * @param array $fields
   * @param array $successors
   *
   * @return $this
   */
  public function addAction(string $id, string $plugin_id, string $label, array $fields, array $successors): Eca {
    if (empty($label)) {
      $label = $id;
    }
    $this->actions[$id] = [
      'plugin' => $plugin_id,
      'label' => $label,
      'fields' => $fields,
      'successors' => $successors,
    ];
    return $this;
  }

  /**
   * @return \Drupal\eca\Entity\Objects\EcaEvent[]
   */
  public function getUsedEvents(): array {
    if ($cached = $this->memoryCache()->get($this->buildCacheId('events'))) {
      return $cached->data;
    }

    $events = [];
    foreach ($this->events as $id => $def) {
      if ($event = $this->getEcaObject('event', $def['plugin'], $id, $def['label'] ?? 'noname', $def['fields'], $def['successors'])) {
        $events[$id] = $event;
      }
    }
    $this->memoryCache()->set($this->buildCacheId('events'), $events, CacheBackendInterface::CACHE_PERMANENT, ['eca.memory_cache:' . $this->id]);
    return $events;
  }

  /**
   * @param \Drupal\eca\Entity\Objects\EcaObject $eca_object
   * @param \Drupal\Component\EventDispatcher\Event $event
   * @param array $context
   *
   * @return \Drupal\eca\Entity\Objects\EcaObject[]
   */
  public function getSuccessors(EcaObject $eca_object, Event $event, array $context): array {
    $successors = [];
    foreach ($eca_object->getSuccessors() as $successor) {
      $context['%successorlabel'] = $successor['label'] ?? 'noname';
      $context['%successorid'] = $successor['id'];
      if ($action = $this->actions[$successor['id']] ?? FALSE) {
        $this->logger()->debug('Check action successor %successorlabel (%successorid) from ECA %ecalabel (%ecaid) for event %event.', $context);
        if ($successorObject = $this->getEcaObject('action', $action['plugin'], $successor['id'], $successor['label'] ?? 'noname', $action['fields'], $action['successors'], $eca_object->getEvent())) {
          if ($this->conditionServices()->assertCondition($event, $successor['condition'], $this->conditions[$successor['condition']] ?? NULL, $context)) {
            $successors[] = $successorObject;
          }
        }
        else {
          $this->logger()->error('Invalid action successor %successorlabel (%successorid) from ECA %ecalabel (%ecaid) for event %event.', $context);
        }
      }
      else if ($gateway = $this->gateways[$successor['id']] ?? FALSE) {
        $this->logger()->debug('Check gateway successor %successorlabel (%successorid) from ECA %ecalabel (%ecaid) for event %event.', $context);
        $successorObject = new EcaGateway($this, $successor['id'], $successor['label'] ?? 'noname', $eca_object->getEvent(), $gateway['type']);
        $successorObject->setSuccessors($gateway['successors']);
        if ($this->conditionServices()->assertCondition($event, $successor['condition'], $this->conditions[$successor['condition']] ?? NULL, $context)) {
          $successors[] = $successorObject;
        }
      }
      else {
        $this->logger()->error('Non existant successor %successorlabel (%successorid) from ECA %ecalabel (%ecaid) for event %event.', $context);
      }
    }
    return $successors;
  }

  /**
   * @param string $type
   * @param string $plugin_id
   * @param string $id
   * @param string $label
   * @param array $fields
   * @param array $successors
   * @param \Drupal\eca\Entity\Objects\EcaEvent|null $event
   *
   * @return \Drupal\eca\Entity\Objects\EcaObject|null
   */
  private function getEcaObject(string $type, string $plugin_id, string $id, string $label, array $fields, array $successors, EcaEvent $event = NULL): ?EcaObject {
    $ecaObject = NULL;
    switch ($type) {
      case 'event':
        try {
          /** @var \Drupal\eca\Plugin\ECA\Event\EventInterface $plugin */
          $plugin = $this->eventPluginManager()->createInstance($plugin_id, $fields);
        }
        catch (PluginException $e) {
          // This can be ignored.
        }
        if (isset($plugin)) {
          $ecaObject = new EcaEvent($this, $id, $label, $plugin);
        }
        break;

      case 'action':
        if ($event !== NULL) {
          try {
            /** @var \Drupal\Core\Action\ActionInterface $plugin */
            $plugin = $this->actionPluginManager()->createInstance($plugin_id, $fields);
          } catch (PluginException $e) {
            // This can be ignored.
          }
          if (isset($plugin)) {
            $ecaObject = new EcaAction($this, $id, $label, $event, $plugin);
          }
        }
        break;

    }
    if ($ecaObject !== NULL) {
      foreach ($fields as $key => $value) {
        $ecaObject->setConfiguration($key, $value);
      }
      $ecaObject->setSuccessors($successors);
      return $ecaObject;
    }
    return NULL;
  }

}
