<?php

namespace Drupal\eca\Entity;

use Drupal\Component\EventDispatcher\Event;
use Drupal\Component\Plugin\ConfigurableInterface;
use Drupal\Component\Plugin\PluginInspectionInterface;
use Drupal\Core\Condition\ConditionInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Logger\LoggerChannel;
use Drupal\eca\Plugin\Action\ActionInterface;
use Drupal\eca\Event\ContentEntityEventInterface;
use Drupal\eca\Event\FormEventInterface;
use Drupal\eca\Token\TokenInterface;

/**
 * A trait for ECA config entities and ECA objects.
 *
 * This trait provides a number of services that we can't inject as dependencies
 * and also provides a set of functions that help to receive a data object from
 * the ECA token service depending on the current context.
 */
trait EcaObjectTrait {

  /**
   * @var \Drupal\Core\Logger\LoggerChannel
   */
  protected LoggerChannel $logger;

  /**
   * @var \Drupal\eca\Token\TokenInterface
   */
  protected TokenInterface $token;

  /**
   * Returns the ECA logger channel as a service.
   *
   * @return \Drupal\Core\Logger\LoggerChannel
   */
  protected function logger(): LoggerChannel {
    if (!isset($this->logger)) {
      $this->logger = \Drupal::service('logger.channel.eca');
    }
    return $this->logger;
  }

  /**
   * Returns the ECA token service.
   *
   * @return \Drupal\eca\Token\TokenInterface
   */
  protected function token(): TokenInterface {
    if (!isset($this->token)) {
      $this->token = \Drupal::service('eca.service.token');
    }
    return $this->token;
  }

  /**
   * Returns the applicable data object for the given plugin.
   *
   * The plugin is either an action or a condition plugin and depending on their
   * type property, this method determines which is the correct data object
   * upon which the action should execute or condition should assert.
   *
   * @param \Drupal\Component\Plugin\PluginInspectionInterface $plugin
   *   The action or condition plugin for which the data object is required.
   *
   * @return \Drupal\Component\EventDispatcher\Event|\Drupal\Core\Entity\EntityInterface|null
   *   The appropriate data object for the given plugin in the current context.
   */
  protected function getObject(PluginInspectionInterface $plugin) {
    $actionType = $plugin->getPluginDefinition()['type'] ?? '';
    switch ($actionType) {
      case NULL:
      case '':
        // The plugin doesn't provide any type declaration, it doesn't require
        // any data object for that matter then.
        return NULL;

      case 'form':
        // The plugin executes upon a form event and this will determine
        // the correct form event to be returned.
        return $this->getFormEvent($plugin);

      case 'system':
      case 'entity':
        // The plugin executes upon an entity and this will determine the
        // correct entity for the current context.
        return $this->getEntity($plugin);

      default:
        // The plugin declares another type, i.e. none of the above. If the
        // given type is an entity type ID and the context provides an entity
        // of that given entity type, this is then the required one and will
        // be returned.
        if (($entity = $this->getEntity($plugin)) && $entity->getEntityTypeId() === $actionType) {
          return $entity;
        }
    }
    return NULL;
  }

  /**
   * Determine the correct entity for the $plugin in the current context.
   *
   * If the plugin is configurable and an entity is being declared as the
   * required one, this will grab that object from the token service and returns
   * it.
   *
   * If the plugin does not request a specific object, this will return the
   * entity from the triggering event for the process in which the given plugin
   * gets executed.
   *
   * @param \Drupal\Component\Plugin\PluginInspectionInterface $plugin
   *   The action or condition plugin for which the data object is required.
   *
   * @return \Drupal\Core\Entity\EntityInterface|null
   *   The required entity if available or NULL otherwise.
   */
  protected function getEntity(PluginInspectionInterface $plugin): ?EntityInterface {
    if ($plugin instanceof ConfigurableInterface && $token = $plugin->getConfiguration()['object'] ?? NULL) {
      $data = $this->token()->getTokenData($token);
      if ($data instanceof EntityInterface) {
        return $data;
      }
    }
    elseif ($plugin instanceof ActionInterface || $plugin instanceof ConditionInterface) {
      $event = $plugin->getEvent();
      if ($event instanceof ContentEntityEventInterface && $entity = $event->getEntity()) {
        $this->token()->addTokenData('entity', $entity);
        return $entity;
      }
    }
    return NULL;
  }

  /**
   * Determine the correct form event for the $plugin in the current context.
   *
   * If the plugin is being executed in the context of a form event, that
   * event will be returned such that the plugin can later retrieve the form
   * and formState objects from that event for further processing.
   *
   * @param \Drupal\Component\Plugin\PluginInspectionInterface $plugin
   *   The action or condition plugin for which the data object is required.
   *
   * @return \Drupal\Component\EventDispatcher\Event|null
   *   The required form event if available or NULL otherwise.
   */
  protected function getFormEvent(PluginInspectionInterface $plugin): ?Event {
    if ($plugin instanceof ActionInterface || $plugin instanceof ConditionInterface) {
      $event = $plugin->getEvent();
      if ($event instanceof FormEventInterface) {
        return $event;
      }
    }
    return NULL;
  }

}
