<?php

namespace Drupal\eca\Entity;

use Drupal\Component\EventDispatcher\Event;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Config\Entity\ConfigEntityStorage;
use Drupal\Core\Entity\EntityTypeInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Storage handler for ECA configurations.
 */
class EcaStorage extends ConfigEntityStorage {

  /**
   * Mapped configurations by event class usage.
   *
   * @var array|null
   */
  protected ?array $configByEvents;

  /**
   * The cache backend for storing prebuilt information.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected CacheBackendInterface $cacheBackend;

  /**
   * {@inheritdoc}
   */
  public static function createInstance(ContainerInterface $container, EntityTypeInterface $entity_type) {
    /** @var static $instance */
    $instance = parent::createInstance($container, $entity_type);
    $instance->setCacheBackend($container->get('cache.default'));
    return $instance;
  }

  /**
   * Loads all ECA configurations that make use of the given event.
   *
   * @param \Drupal\Component\EventDispatcher\Event $event
   *   The event object.
   *
   * @return \Drupal\eca\Entity\Eca[]
   *   The configurations, keyed by entity ID.
   */
  public function loadByEvent(Event $event): array {
    if (!isset($this->configByEvents)) {
      $cid = 'eca:storage:events';
      if ($cached = $this->cacheBackend->get($cid)) {
        $this->configByEvents = $cached->data;
      }
      else {
        $this->configByEvents = [];
        $entities = $this->loadMultiple();
        // Sort the configurations by weight and label.
        uasort($entities, [$this->entityType->getClass(), 'sort']);
        /** @var \Drupal\eca\Entity\Eca $eca */
        foreach ($entities as $eca) {
          if (!$eca->status()) {
            continue;
          }
          foreach ($eca->getUsedEvents() as $ecaEvent) {
            $used_class = $ecaEvent->getPlugin()->drupalEventClass();
            if (!isset($this->configByEvents[$used_class])) {
              $this->configByEvents[$used_class] = [$eca->id()];
            }
            elseif (!in_array($eca->id(), $this->configByEvents[$used_class], TRUE)) {
              $this->configByEvents[$used_class][] = $eca->id();
            }
          }
        }
        $this->cacheBackend->set($cid, $this->configByEvents, CacheBackendInterface::CACHE_PERMANENT, ['config:eca_list']);
      }
    }
    $class = get_class($event);
    if (empty($this->configByEvents[$class])) {
      return [];
    }
    return $this->loadMultiple($this->configByEvents[$class]);
  }

  /**
   * {@inheritdoc}
   */
  public function resetCache(array $ids = NULL): void {
    $this->configByEvents = NULL;
    parent::resetCache($ids);
  }

  /**
   * Set the cache backend for storing prebuilt information.
   *
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache_backend
   *   The cache backend.
   */
  public function setCacheBackend(CacheBackendInterface $cache_backend): void {
    $this->cacheBackend = $cache_backend;
  }

}
