<?php

namespace Drupal\eca\Plugin\Condition;

use Drupal\Core\Form\FormStateInterface;
use Drupal\eca\Service\Conditions;
use Drupal\eca\Plugin\OptionsInterface;

/**
 * Base class for ECA condition plugins to compare strings.
 */
abstract class StringComparisonBase extends ConditionBase implements OptionsInterface {

  public const COMPARE_EQUALS = 'equal';
  public const COMPARE_BEGINS_WITH = 'beginswith';
  public const COMPARE_ENDS_WITH = 'endswith';
  public const COMPARE_CONTAINS = 'contains';

  /**
   * Get the first string value for comparison.
   *
   * @return string
   */
  abstract protected function getFirstValue(): string;

  /**
   * Get the second string value for comparison.
   *
   * @return string
   */
  abstract protected function getSecondValue(): string;

  /**
   * @return bool
   */
  protected function caseSensitive(): bool {
    return $this->configuration['case'] === Conditions::OPTION_YES;
  }

  /**
   * {@inheritdoc}
   */
  final public function evaluate(): bool {
    $value1 = $this->tokenServices->replace($this->getFirstValue());
    $value2 = $this->tokenServices->replace($this->getSecondValue());
    if (!$this->caseSensitive()) {
      $value1 = strtolower($value1);
      $value2 = strtolower($value2);
    }
    switch ($this->configuration['type']) {
      case self::COMPARE_EQUALS:
        return $value1 === $value2;

      case self::COMPARE_BEGINS_WITH:
        return strpos($value1, $value2) === 0;

      case self::COMPARE_ENDS_WITH:
        return strpos($value1, $value2) === (strlen($value1) - strlen($value2));

      case self::COMPARE_CONTAINS:
        return strpos($value1, $value2) !== FALSE;

    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration(): array {
    return [
        'case' => FALSE,
        'type' => self::COMPARE_EQUALS,
      ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form['case'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Case sensitive comparison'),
      '#default_value' => $this->configuration['case'],
    ];
    $form['type'] = [
      '#type' => 'select',
      '#title' => $this->t('Comparison type'),
      '#default_value' => $this->configuration['type'],
      '#options' => $this->getOptions('type'),
    ];
    return parent::buildConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state): void {
    $this->configuration['case'] = (bool) $form_state->getValue('case');
    $this->configuration['type'] = $form_state->getValue('type');
    parent::submitConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function getOptions(string $id): ?array {
    if ($id === 'type') {
      return [
        self::COMPARE_EQUALS => $this->t('equals'),
        self::COMPARE_BEGINS_WITH => $this->t('begins with'),
        self::COMPARE_ENDS_WITH => $this->t('ends with'),
        self::COMPARE_CONTAINS => $this->t('contains'),
      ];
    }
    return NULL;
  }

}
