<?php

namespace Drupal\eca\Service;

use Drupal\Component\Plugin\PluginInspectionInterface;
use Drupal\Core\Form\FormState;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\eca\Plugin\OptionsInterface;

/**
 * Trait for ECA modeller, condition and action services.
 */
trait ServiceTrait {

  /**
   * Helper function to sort plugins by their label.
   *
   * @param \Drupal\Component\Plugin\PluginInspectionInterface[] $plugins
   *   The list of plugin to be sorted.
   */
  public function sortPlugins(array &$plugins): void {
    usort($plugins, static function($p1, $p2) {
      $l1 = (string) $p1->getPluginDefinition()['label'];
      $l2 = (string) $p2->getPluginDefinition()['label'];
      if ($l1 < $l2) {
        return -1;
      }
      if ($l1 > $l2) {
        return 1;
      }
      return 0;
    });
  }

  /**
   * Helper function to prepare a config field for actions and conditions.
   *
   * @param array $fields
   *   The array to which the fields should be added.
   * @param array $config
   *   The array received from defaultConfiguraiton of actions and conditions.
   * @param \Drupal\Component\Plugin\PluginInspectionInterface $plugin
   *   The plugin to which the field belong to.
   */
  protected function prepareConfigFields(array &$fields, array $config, PluginInspectionInterface $plugin): void {
    foreach ($config as $key => $value) {
      $label = NULL;
      if ($plugin instanceof PluginFormInterface) {
        $form = [];
        $form_state = new FormState();
        $form = $plugin->buildConfigurationForm($form, $form_state);
        if (isset($form[$key]['#title'])) {
          $label = (string) $form[$key]['#title'];
        }
      }
      if ($label === NULL) {
        $labelParts = explode('_', $key);
        $labelParts[0] = strtoupper($labelParts[0][0]) . substr($labelParts[0], 1);
        $label = implode(' ', $labelParts);
      }
      if (is_bool($value)) {
        $fields[] = $this->checkbox($key, $label, $value);
        continue;
      }
      if (is_array($value)) {
        $value = implode(',', $value);
      }
      elseif ($plugin instanceof OptionsInterface && $options = $plugin->getOptions($key)) {
        $fields[] = $this->optionsField($key, $label, $options, $value);
        continue;
      }
      $fields[] = [
        'name' => $key,
        'label' => $label,
        'type' => 'String',
        'value' => $value,
      ];
    }
  }

  /**
   * Prepares a field with options as a drop-down.
   *
   * @param string $name
   *   The field name.
   * @param string $label
   *   The field label.
   * @param array $options
   *   Key/value list of available options.
   * @param string $value
   *   The default value for the field.
   *
   * @return array
   *   Prepared option field.
   */
  protected function optionsField(string $name, string $label, array $options, string $value): array {
    $choices = [];
    foreach ($options as $optionValue => $optionName) {
      $choices[] = [
        'name' => $optionName,
        'value' => $optionValue,
      ];
    }
    return [
      'name' => $name,
      'label' => $label,
      'type' => 'Dropdown',
      'value' => $value,
      'extras' => [
        'choices' => $choices,
      ],
    ];
  }

  /**
   * Prepares a field as a checkbox.
   *
   * @param $name
   *   The field name.
   * @param $label
   *   The field label.
   * @param $value
   *   The default value for the field.
   *
   * @return array
   *   Prepared checkbox field.
   */
  protected function checkbox($name, $label, $value): array {
    return [
      'name' => $name,
      'label' => $label,
      'type' => 'Dropdown',
      'value' => $value ? Conditions::OPTION_YES : Conditions::OPTION_NO,
      'extras' => [
        'choices' => [
          [
            'name' => 'no',
            'value' => Conditions::OPTION_NO,
          ],
          [
            'name' => 'yes',
            'value' => Conditions::OPTION_YES,
          ],
        ],
      ],
    ];
  }

}
