<?php

namespace Drupal\eca\Token;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Utility\Token;
use Drupal\Core\Render\BubbleableMetadata;

/**
 * A trait for ECA-specific token service decorators.
 *
 * The token service is being extended this way to support collection of data
 * which may be added on runtime, plus it allows the usage of aliases in order
 * to use multiple data sets of the same type, e.g. you can use both the
 * currently logged in user and the author user of a node to replace values
 * in a given text.
 *
 * @see \Drupal\eca\Token\TokenInterface
 */
trait TokenDecoratorTrait {

  /**
   * The decorated token service.
   *
   * IMPORTANT: Do not define the variable type here because the same variable
   * name is used in ContribToken but it has a different namespace there, which
   * leads to PHP error at runtime.
   *
   * @var \Drupal\Core\Utility\Token
   */
  protected $token;

  /**
   * An array of currently hold token data.
   *
   * @var array
   */
  protected array $data = [];

  /**
   * A map of aliases to their token type.
   *
   * @var array
   */
  protected array $aliases = [];

  /**
   * Set the token service that is being decorated by this service.
   *
   * @param \Drupal\Core\Utility\Token $token
   *   The token service to decorate.
   */
  public function setDecoratedToken(Token $token): void {
    $this->token = $token;
  }

  /**
   * {@inheritdoc}
   */
  public function addTokenData(string $key, $data): TokenInterface {
    $tokenType = $this->getTokenType($data);
    if ($tokenType === NULL) {
      // No token type available, consider this being arbitrary data.
      $this->data['dto'][$key] = $data;
    }
    else {
      $this->data[$key] = $data;
      if ($tokenType !== $key) {
        $this->aliases[$key] = $tokenType;
      }
    }
    return $this;
  }

  /**
   * Gets the token type of data object if possible.
   *
   * @param $value
   *   Data object for which to determine the token type.
   *
   * @return string|null
   *   The token type if available, NULL otherwise.
   */
  protected function getTokenType($value): ?string {
    $aliasFor = NULL;
    if ($value instanceof EntityInterface) {
      if (\Drupal::hasService('token.entity_mapper')) {
        /** @var \Drupal\token\TokenEntityMapperInterface $token_entity_mapper */
        $token_entity_mapper = \Drupal::service('token.entity_mapper');
        $aliasFor = $token_entity_mapper->getTokenTypeForEntityType($value->getEntityTypeId(), TRUE);
      }
      else if ($definition = \Drupal::entityTypeManager()->getDefinition($value->getEntityTypeId())) {
        $aliasFor = $definition->get('token_type') ?: $value->getEntityTypeId();
      }
      else {
        $aliasFor = $value->getEntityTypeId();
      }
    }
    return $aliasFor;
  }

  /**
   * {@inheritdoc}
   */
  public function getTokenData(string $key) {
    return $this->data[$key] ?? $this->replace('[' . $key . ']');
  }

  /**
   * {@inheritdoc}
   */
  public function clearTokenData(): void {
    $this->data = [];
    $this->aliases = [];
  }

  /**
   * {@inheritdoc}
   */
  public function generate($type, array $tokens, array $data, array $options, BubbleableMetadata $bubbleable_metadata) {
    // Merge in previously added data (if any), but data which comes directly
    // from the given input always wins.
    $data += $this->data;
    if (isset($this->aliases[$type])) {
      // Given $type input is an alias, use its real type.
      $alias = $type;
      $type = $this->aliases[$alias];
      if (isset($data[$alias])) {
        // Make sure, that the assigned data for the given alias will be used
        // when generating the token replacement values.
        $data[$type] = $data[$alias];
      }
    }
    // Now that we have mapped a possibly given alias to its type, we can let
    // the decorated token service do its original job. That passthrough will
    // not overwrite any other aliased data, because the returned token
    // replacements are keyed by their "raw" original token input, and that
    // always includes the alias as prefix (or just the type when not working
    // with aliases at all).
    return $this->token->generate($type, $tokens, $data, $options, $bubbleable_metadata);
  }

  /**
   * {@inheritdoc}
   */
  public function replace($text, array $data = [], array $options = [], BubbleableMetadata $bubbleable_metadata = NULL) {
    // @todo Prepare arguments, see
    // https://www.drupal.org/project/eca/issues/3232083.
    // Replacement of aliased tokens can only work within the scope of this
    // decorator. Thus we call it on its own.
    $text = parent::replace($text, $data, $options, $bubbleable_metadata);

    // Either the class of this decorator inherits from the Core token service
    // or from the Contrib token service (if available). Just in case we
    // actually received a decorated service that differs from these two
    // implementation variants, give it a chance to execute its own logic.
    if (get_parent_class($this) !== get_class($this->token)) {
      $text = $this->token->replace($text, $data, $options, $bubbleable_metadata);
    }
    return $text;
  }

  /**
   * {@inheritdoc}
   */
  public function scan($text) {
    return $this->token->scan($text);
  }

  /**
   * {@inheritdoc}
   */
  public function findWithPrefix(array $tokens, $prefix, $delimiter = ':') {
    return $this->token->findWithPrefix($tokens, $prefix, $delimiter);
  }

  /**
   * {@inheritdoc}
   */
  public function getInfo(): array {
    return $this->token->getInfo();
  }

  /**
   * {@inheritdoc}
   */
  public function setInfo(array $tokens): void {
    $this->token->setInfo($tokens);
  }

  /**
   * {@inheritdoc}
   */
  public function resetInfo(): void {
    $this->token->resetInfo();
  }

}
