<?php

namespace Drupal\eca\Token;

use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Render\BubbleableMetadata;
use Drupal\Core\Utility\Token;

/**
 * Wrapper class for ECA-related token services.
 *
 * This is an intermediary that should be used when working with token aliases
 * and when adding token data on runtime (TokenInterface::addTokenData()).
 * The reason why this wrapper exists, is because we use the decorator pattern
 * to extend the existing 'token' service. By going this extension path, we
 * want to make sure, that we are always able to make use of our method, and
 * still allow other modules to also extend the token service any further with
 * their own decorators.
 */
class TokenServices implements TokenInterface {

  /**
   * The ECA-specific decorator instance of the token service.
   *
   * @var \Drupal\eca\Token\TokenInterface
   */
  protected TokenInterface $decorator;

  /**
   * The final instance of the token service.
   *
   * @var \Drupal\Core\Utility\Token
   */
  protected Token $token;

  /**
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected LoggerChannelInterface $logger;

  /**
   * The TokenServices constructor.
   *
   * @param \Drupal\eca\Token\TokenInterface $eca_decorator
   *   The ECA-specific decorator instance of the token service.
   * @param \Drupal\Core\Utility\Token $token
   *   The final instance of the token service.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger
   */
  public function __construct(TokenInterface $eca_decorator, Token $token, LoggerChannelInterface $logger) {
    $this->decorator = $eca_decorator;
    $this->token = $token;
    $this->logger = $logger;
  }

  /**
   * {@inheritdoc}
   */
  public function addTokenData(string $key, $data): TokenInterface {
    $this->decorator->addTokenData($key, $data);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getTokenData(string $key) {
    return $this->decorator->getTokenData($key);
  }

  /**
   * {@inheritdoc}
   */
  public function clearTokenData(): void {
    $this->decorator->clearTokenData();
  }

  /**
   * {@inheritdoc}
   */
  public function generate($type, array $tokens, array $data, array $options, BubbleableMetadata $bubbleable_metadata): array {
    return $this->token->generate($type, $tokens, $data, $options, $bubbleable_metadata);
  }

  /**
   * {@inheritdoc}
   */
  public function replace($text, array $data = [], array $options = [], BubbleableMetadata $bubbleable_metadata = NULL): string {
    return $this->token->replace($text, $data, $options, $bubbleable_metadata);
  }

  /**
   * {@inheritdoc}
   */
  public function scan($text): array {
    return $this->token->scan($text);
  }

  /**
   * {@inheritdoc}
   */
  public function findWithPrefix(array $tokens, $prefix, $delimiter = ':'): array {
    return $this->token->findWithPrefix($tokens, $prefix, $delimiter);
  }

  /**
   * {@inheritdoc}
   */
  public function getInfo(): array {
    return $this->token->getInfo();
  }

  /**
   * {@inheritdoc}
   */
  public function setInfo(array $tokens): void {
    $this->token->setInfo($tokens);
  }

  /**
   * {@inheritdoc}
   */
  public function resetInfo(): void {
    $this->token->resetInfo();
  }

}
